"""Declares :class:`ConnectionManager`."""
import os

from sqlalchemy.ext.asyncio import AsyncSession
from unimatrix.lib import rdbms

from .connection import Connection


class ConnectionManager:
    """Manages the connections for an application."""
    __module__ = 'unimatrix.ext.rdbms'

    @property
    def config(self) -> dict:
        """A mapping of aliases to :class:`Connection` objects."""
        return self.__connections

    @config.setter
    def config(self, value: dict) -> dict:
        """A mapping of aliases to :class:`Connection` objects."""
        self.__connections = value

    def __init__(self, environ=None):
        self.__connections = {
            name: Connection(opts.dsn)
            for name, opts in dict.items(
                rdbms.load_config(env=environ or os.environ)
            )
        }

    def _connection_factory(self, opts: dict) -> Connection:
        opts = rdbms.load_config(opts)
        return Connection(opts['self'].dsn)

    def add(self, name: str, opts: dict) -> None:
        """Add a new connection using the given parameters"""
        self.__connections[name] = self._connection_factory(opts)

    def get(self, name: str = 'self') -> Connection:
        """Return the named connection `name`."""
        return self.__connections[name]

    async def clear(self):
        """Disconnect all databases and remove them from the internal
        registry.
        """
        await self.disconnect()
        self.__connections = {}

    async def connect(self, *args, **kwargs):
        """Connect all database connections that are specified."""
        connected = []
        for name, connection in dict.items(self.__connections):
            try:
                await connection.connect(*args, **kwargs)
                connected.append(connection)
            except Exception as e:
                for active in connected:
                    await active.disconnect()
                raise

    async def disconnect(self):
        """Disconnect all database connections that are specified."""
        for name, connection in dict.items(self.__connections):
            try:
                await connection.disconnect()
            except Exception as e:
                pass


#: The connections that could be automatically determined from the application
#: environment and predefined locations.
connections = ConnectionManager()


def get(name: str) -> Connection:
    """Return the named connection `name`."""
    return connections.get(name)


def session(name: str) -> AsyncSession:
    """Create a new :class:`sqlalchemy.ext.asyncio.AsyncSession`
    instance for the named connection `name`.
    """
    connection = get(name)
    if not connection.is_connected():
        raise RuntimeError(f"Connection '{name}' is not established.")
    return connection.get_session()



# pylint: skip-file
import importlib

import sqlalchemy.ext.declarative
from sqlalchemy.sql.schema import MetaData
from unimatrix.conf import settings


def declarative_base(cls=None, class_factory=None, settings=None, apps=None): # pragma: no cover
    """Like :func:`sqlalchemy.ext.declarative.declarative_base()`, but
    returns a tuple containing the metadata objects for all applications
    that were specified in :attr:`unimatrix.conf.settings.INSTALLED_APPS`.

    Each application must have a submodule/subpackage named ``orm``, which
    exposed a ``metadata`` attribute.

    The list of installed apps may also be provided using the `apps`
    parameter.

    The resulting list of metadata objects is used to create Alembic
    migration scripts.

    Optionally, the `cls` argument may be provided to use a pre-existing
    declarative base class instead. This allows the :func:`declarative_base`
    function to be used with legacy models.

    By default, :func:`sqlalchemy.ext.declarative.declarative_base()` is
    used to create the base class, but another factory may be specified using
    the `class_factory` argument.
    """
    class_factory = class_factory or sqlalchemy.ext.declarative.declarative_base
    Base = cls or (class_factory)()
    target_metadata = []
    apps = apps or getattr(settings, 'INSTALLED_APPS', [])
    if apps:
        for app_qualname in apps:
            orm_qualname = "%s.orm" % app_qualname
            try:
                orm = importlib.import_module(orm_qualname)
            except ImportError as e:
                # Fail silently - assume that the installed application does
                # not declare an ORM. We do want to catch other ImportError
                # exceptions tho. This is quicky flaky but it works for now.
                if e.args[0] != ("No module named '%s'" %  orm_qualname):
                    raise
                continue
            if isinstance(orm.metadata, MetaData):
                target_metadata.append(orm.metadata)
            elif isinstance(orm.metadata, (list, tuple)):
                target_metadata.extend(orm.metadata)
            else:
                raise TypeError(
                    "%s.metadata is of invalid type." % orm_qualname)

    target_metadata.append(Base.metadata)
    return Base, target_metadata




# pylint: skip-file
import contextlib

from sqlalchemy.ext.asyncio import AsyncSession


class TestSession:
    __module__ = 'unimatrix.ext.rdbms.testsession'
    __test__ = False

    def __init__(self, session):
        assert session is not None # nosec
        assert isinstance(session, AsyncSession), session # nosec
        self.__session = session
        self.__transaction = None

    @contextlib.asynccontextmanager
    async def begin(self):
        yield

    async def close(self):
        # Do nothing, since we control the closing of the
        # session.
        pass

    async def destroy(self):
        if self.__transaction is not None:
            await self.__transaction.rollback()
        await self.__session.close()
        self.__session = self.__transaction = None

    def __getattr__(self, attname):
        assert self.__session is not None # nosec
        return getattr(self.__session, attname)

    def __call__(self, *args, **kwargs):
        return self

    async def __aenter__(self):
        if not self.__transaction:
            self.__transaction = await self.__session.begin()
        return self

    async def __aexit__(self, cls, exception, traceback):
        pass



# pylint: skip-file
from unimatrix.lib.test import TEST_STAGE

from .fixtures import setup_databases


def pytest_sessionstart(*args, **kwargs) -> None:
    """If :attr:`unimatrix.lib.test.TEST_STAGE` is ``system``, and database
    connections are configured, create a test database for each connection.
    """
    if TEST_STAGE != 'system':
        return
    setup_databases()



# pylint: skip-file
import asyncio

import pytest

from unimatrix.ext import rdbms
from ..connection import Connection


@pytest.mark.asyncio
async def test_max_connections_raises(postgresql):
    engines = []
    for i in range(10):
        engines.append(Connection(postgresql.dsn))
    await asyncio.gather(*[x.connect(debug=True) for x in engines])

    futures = []
    for e in engines:
        session = e.get_session()

        async def f():
            await session.execute("SELECT 1")
            await asyncio.sleep(2)
            await session.close()

        futures.append(f())

    await asyncio.gather(*futures)
    raise Exception



"""Declares :class:`Repository`."""
import typing

import sqlalchemy
from sqlalchemy.engine import Result
from sqlalchemy.exc import NoResultFound
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.sql.dml import Delete
from sqlalchemy.sql.dml import Update
from sqlalchemy.sql.selectable import Select
from unimatrix.ext import rdbms


EXECUTABLE = typing.Union[Delete, Update, Select]
RESULT = typing.Union[Result]


class Repository:
    """Integrates the :term:`Repository Layer` with relational database systems
    using the :mod:`sqlalchemy` package. Implements :class:`ddd.Repository`.
    """
    #: Specifies the connection that is used by the repository implementation,
    #: as specified using the :mod:`unimatrix.ext.rdbms` framework.
    alias: str = 'self'

    async def get_tuple(self, session: AsyncSession, query: EXECUTABLE):
        """Return a single tuple from the given query. The query is expected
        to yield exactly one tuple. Raises :class:`ddd.DoesNotExist` if there
        is no tuple matched by the predicate.
        """
        result = await self.execute(session, query)
        try:
            return result.one()
        except NoResultFound:
            raise self.DoesNotExist

    async def execute(self, session: AsyncSession, query: EXECUTABLE) -> Result:
        """Execute the given `query` and return the result. Does not evaluate
        the result.
        """
        return await session.execute(query)

    async def nextval(self, session: AsyncSession, name: str) -> int:
        """Return the next value of sequence `name`."""
        result = await self.execute(session, sqlalchemy.func.nextval(name))
        return result.scalars().one()

    async def persist_declarative(self,
        session: AsyncSession,
        dao: object,
        merge: bool = False,
        flush: bool = False
    ) -> 'dao':
        """Persist a declarative SQLAlchemy object."""
        if not merge:
            session.add(dao)
        else:
            await session.merge(dao)
        if flush:
            await session.flush()
        return dao

    def begin(self) -> AsyncSession:
        """Return the :class:`sqlalchemy.ext.asyncio.AsyncSession` instance
        that is used for this repository.
        """
        return rdbms.session(self.alias)



# pylint: skip-file
import asyncio
import tempfile

import dsnparse
import pytest
from unimatrix.lib.rdbms.config import encode_dsn
from unimatrix.lib.rdbms.config import parse_environment

from ..connection import Connection


@pytest.fixture
def event_loop():
    return asyncio.get_event_loop()


@pytest.fixture
def sqlite():
    _, fn = tempfile.mkstemp()
    return Connection(
        encode_dsn(
            parse_environment({
                'DB_ENGINE': "sqlite",
                'DB_NAME': fn
            })
        )
    )


@pytest.fixture
def postgresql():
    return Connection(
        encode_dsn(
            parse_environment({
                'DB_ENGINE': 'postgresql',
                'DB_HOST': 'localhost',
                'DB_NAME': 'rdbms',
                'DB_USERNAME': 'rdbms',
                'DB_PASSWORD': 'rdbms',
            })
        )
    )



# pylint: skip-file
import unittest

import pytest

from unimatrix.ext import rdbms
from unimatrix.ext.rdbms.testsession import TestSession


@pytest.mark.asyncio
@pytest.mark.usefixtures('rdbms_transaction')
async def test_session_is_wrapped():
    connection = rdbms.get('self')
    session = connection.get_session()
    assert type(session).__name__ == 'TestSession'


@pytest.mark.asyncio
@pytest.mark.usefixtures('rdbms_transaction')
async def test_generic_statements():
    connection = rdbms.get('self')
    async with connection.get_session() as session:
        async with session.begin():
            await session.execute("SELECT 1")



"""Declares :class:`Connection`."""
import copy
import threading
import typing
import os

import dsnparse
from sqlalchemy import create_engine
from sqlalchemy.engine import Engine
from sqlalchemy.ext.asyncio import create_async_engine
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import sessionmaker
from sqlalchemy.pool import QueuePool
from unimatrix.conf import settings


class Connection:
    """A wrapper around :class:`databases.Database`."""
    async_schemes = {
        'mysql'     : 'mysql+aiomysql',
        'postgresql': 'postgresql+asyncpg',
        'sqlite'    : 'sqlite+aiosqlite'
    }

    @property
    def db_name(self):
        return str.lstrip(self.__opts.get('path') or '', '/') or None

    @property
    def dsn(self) -> str:
        """Return the Data Source Name (DSN) for this connection."""
        return self.__dsn.geturl()

    @property
    def engine(self) -> typing.Optional[Engine]:
        if not hasattr(self.__local, 'engine'):
            self.__local.engine = None
        return self.__local.engine

    @engine.setter
    def engine(self, value):
        self.__local.engine = value

    @property
    def session_factory(self):
        return self.__local.session_factory

    @session_factory.setter
    def session_factory(self, value):
        self.__local.session_factory = value

    def __init__(self, dsn):
        self.__local = threading.local()
        self.__local.engine = None
        self.__local.session_factory = None
        self.__dsn = dsn
        self.__opts = self.__dsn.parse(self.__dsn.geturl())
        self.__session = None

    def as_test(self):
        """Return a new instance with its connection options configured
        for testing.
        """
        return type(self)(self.__dsn.test)

    def _get_server_engine(self):
        if self.__dsn.scheme == 'sqlite':
            return
        engine = create_engine(self.__dsn.admin.geturl(), echo=True)
        session = sessionmaker(bind=engine)()
        if self.__dsn.scheme == 'postgresql':
            session.connection()\
                .connection.set_isolation_level(0)
        return session

    def create_database(self):
        """Creates the database as specified by the connection parameters."""
        session = self._get_server_engine()
        if session is not None:
            session.execute(f"CREATE DATABASE {self.db_name}")

    def drop_database(self):
        """Drops the database as specified by the connection parameters."""
        session = self._get_server_engine()
        if session is not None:
            session.execute(f"DROP DATABASE {self.db_name}")

    def begin(self):
        """Begin a transaction without any session management using the
        underlying database connection.
        """
        return self.engine.begin()

    def clone(self):
        """Clones the connection."""
        assert not self.is_connected() # nosec
        return copy.deepcopy(self)

    async def connect(self, debug: bool = False, *args, **kwargs) -> None:
        """Connect to the database server."""
        if self.engine is not None:
            raise RuntimeError("Connection already established.")
        self.engine = create_async_engine(
            self.get_dsn_async(),
            connect_args={"timeout": settings.DB_TIMEOUT},
            echo=debug,
            pool_size=settings.DB_MAX_CONNECTIONS,
            pool_timeout=settings.DB_TIMEOUT,
            poolclass=QueuePool,
        )
        self.session_factory = sessionmaker(
            self.engine,
            expire_on_commit=False,
            class_=AsyncSession
        )

    async def disconnect(self, *args, **kwargs) -> None:
        """Disconnect from the database server."""
        if self.engine is None:
            return
        await self.engine.dispose()
        self.engine = None
        self.session_factory = None

    def get_session(self, *args, **kwargs) -> AsyncSession:
        """Establish a new session."""
        if self.__session is not None:
            return self.__session

        return self.session_factory(*args, **kwargs)

    def get_dsn_async(self) -> str:
        """Return a string containing the Data Source Name (DSN)
        for asynchronous connections.
        """
        return str.replace(
            self.__dsn.geturl(),
            self.__dsn.scheme,
            self.async_schemes[self.__dsn.scheme]
        )

    def is_connected(self) -> bool:
        """Return a boolean indicating if the connection
        is established.
        """
        return self.engine is not None



"""Declares :class:`QueryRunner`."""
from .connectionmanager import session as session_factory


class QueryRunner:
    """Provides an interface to query a relational database."""
    db_alias: str = 'self'

    async def execute(self, query, scalar=False):
        """Execute `query` and return the result."""
        result = await self.session.execute(query)
        if scalar:
            result = result.scalar()
        return result

    async def __aenter__(self):
        self.session=session_factory(self.db_alias)
        return self

    async def __aexit__(self, cls, exception, traceback):
        # Note that there is no commit or rollback here, since
        # QueryRunner does not mutate data.
        await self.session.close()



# pylint: skip-file
from .connectionmanager import connections
from .connection import Connection
from .connectionmanager import get
from .connectionmanager import session
from .connectionmanager import ConnectionManager
from .declarative import declarative_base
from .queryrunner import QueryRunner
from .repository import Repository


__all__ = [
    'Connection',
    'ConnectionManager',
    'QueryRunner',
    'Repository',
    'add',
    'connect',
    'declarative_base',
    'disconnect',
    'get',
    'session',
    'setup_databases',
]


def add(name: str, **opts) -> None:
    """Add a new connection using the given parameters"""
    connections.add(name, opts)


async def connect(*args, **kwargs) -> None:
    """Connect all database connections that are specified in the default
    connection manager.
    """
    return await connections.connect(*args, **kwargs)


async def disconnect() -> None:
    """Disconnect all database connections that are specified in the default
    connection manager.
    """
    return await connections.disconnect()



# pylint: skip-file
import asyncio
import unittest

import sqlalchemy

from .. import connections


class BaseTestCase(unittest.TestCase):

    def setUp(self):
        try:
            self.loop = asyncio.get_event_loop()
        except RuntimeError:
            self.loop = asyncio.new_event_loop()
            asyncio.set_event_loop(self.loop)
        self.metadata = sqlalchemy.MetaData()

    def tearDown(self):
        self.run_async(connections.disconnect)

    def assertCanConnect(self, name: str):
        connection = connections.get(name)
        self.run_async(connection.connect)

        handle = self.run_async(connection.begin)
        self.run_async(handle.close)

    def add_connection(self, name, opts):
        connections.add(name, opts)

    def run_async(self, func, *args, **kwargs):
        return self.loop.run_until_complete(func(*args, **kwargs))

    def connect_databases(self):
        self.run_async(connections.connect, debug=True)

    def create_all(self, connection):
        connection.create_all(self.metadata)

    def get_connection(self, name='self'):
        return connections.get(name)
